/* $Id: EvaluationVisitor.java,v 1.6 2006/03/26 20:07:15 gsson Exp $ */

/*
 * Copyright (c) 2006 Henrik Gustafsson <henrik.gustafsson@fnord.se>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

package se.fnord.ecalc.parser;

public class EvaluationVisitor implements ExpressionParserVisitor {

	public IValue visit(SimpleNode node, ExpressionContext o) throws ExpressionException {
		throw new IllegalArgumentException();
	}

	public IValue visit(ASTStart node, ExpressionContext o) throws ExpressionException {
		if (node.jjtGetNumChildren() == 1) {
			return ((SimpleNode) node.jjtGetChild(0)).accept(this, o).ensureType(o.arithmeticType);
		}
		throw new IllegalArgumentException();
	}

	public IValue visit(ASTExpression node, ExpressionContext o) throws ExpressionException {
		if (node.jjtGetNumChildren() == 1) {
			return ((SimpleNode) node.jjtGetChild(0)).accept(this, o);
		}
		throw new IllegalArgumentException();
	}

	public IValue visit(ASTAssign node, ExpressionContext o) throws ExpressionException {
		if (node.jjtGetNumChildren() == 1) {
			return ((SimpleNode) node.jjtGetChild(0)).accept(this, o);
		} else if (node.jjtGetNumChildren() > 1) {
			IValue rhs = ((SimpleNode) node.jjtGetChild(node.jjtGetNumChildren() - 1)).accept(this, o);
			for (int i = node.jjtGetNumChildren() - 2; i >= 0; i--) {
				SimpleNode n = (SimpleNode)node.jjtGetChild(i);
				if (n instanceof ASTVariable)
					((Literal)n.accept(this, o)).fromIValue(rhs);
				else
					throw new ExpressionException("%se.fnord.ecalc.exceptions.expressionexception.assignerror");
			}
			return rhs;
		}
		throw new IllegalArgumentException();
	}

	public IValue visit(ASTBinaryOperator node, ExpressionContext o) throws ExpressionException {
		
		Literal lhs = ((SimpleNode) node.jjtGetChild(0)).accept(this, o).ensureType(o.arithmeticType);

		for (int i = 1; i < node.jjtGetNumChildren(); i++) {
			IValue rhs = ((SimpleNode) node.jjtGetChild(i)).accept(this, o).ensureType(o.arithmeticType);
			switch(((Integer)node.operators.get(i - 1)).intValue()) {
			case ASTBinaryOperator.ADDITION:
				lhs.add(rhs);
				break;
			case ASTBinaryOperator.SUBTRACTION:
				lhs.subtract(rhs);
				break;
			case ASTBinaryOperator.MULTIPLICATION:
				lhs.multiply(rhs);
				break;
			case ASTBinaryOperator.DIVISION:
				lhs.divide(rhs);
				break;
			case ASTBinaryOperator.MODULUS:
				lhs.modulus(rhs);
				break;
			case ASTBinaryOperator.POWER:
				lhs.power(rhs);
				break;
			case ASTBinaryOperator.SHL:
				lhs.shift_left(rhs);
				break;
			case ASTBinaryOperator.SHR:
				lhs.shift_right(rhs);
				break;
			case ASTBinaryOperator.AND:
				lhs.bitwise_and(rhs);
				break;
			case ASTBinaryOperator.OR:
				lhs.bitwise_or(rhs);
				break;
			default:
				throw new IllegalArgumentException();
			}
		}
		return lhs;
	}

	public IValue visit(ASTUnary node, ExpressionContext o) throws ExpressionException {
		if (node.jjtGetNumChildren() == 1) {
			IValue v = ((SimpleNode) node.jjtGetChild(0)).accept(this, o);
			if (node.operator == ASTUnary.NEGATE)
				return v.ensureType(o.arithmeticType).negate();
			else
				return v;
		}
		throw new IllegalArgumentException();
	}

	public IValue visit(ASTLiteral node, ExpressionContext o) throws ExpressionException {
		return new Literal(node.value);
	}

	public IValue visit(ASTVariable node, ExpressionContext o) throws ExpressionException {
		IValue v = o.getSymbol(node.name);
		return v;
	}

	public IValue visit(ASTFunction node, ExpressionContext o) throws ExpressionException {
		
		IFunction f = o.getFunction(node.name);
		IValue[] v = new IValue[node.jjtGetNumChildren()];
		for (int i = 0; i < node.jjtGetNumChildren(); i++) {
			v[i] = ((SimpleNode) node.jjtGetChild(i)).accept(this, o).ensureType(o.arithmeticType);
		}
		return f.execute(v, o);
	}

}
